/* liste Primzahlen aus einer Primzahlendatei auf */

#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>

#define ZEILE 79  /* die letzte Spalte bleibt frei */

typedef struct
  { unsigned long int oben;
    unsigned long int unten;
  } ULTRALANG;

void gibaus(ULTRALANG);
ULTRALANG uladd(ULTRALANG, unsigned long int);
unsigned long int uldiff(ULTRALANG, ULTRALANG);

main(int argc, char *argv[])
{ char *primfile, *zeiger;
  unsigned long int i, anzahl = ~0ul;
  unsigned int wort;
  int j;
  ULTRALANG start = {0, 0}, offset;
  FILE *datei;

  puts("PRIMLIST 1.0 - Primzahlenausgabe (c) 1995 Heinz Repp");
  switch (argc)
  { case 4:
      zeiger = argv[2];
      i = 0; /* Zwischenspeicher fr Bit 16-31 */
      while (isdigit(*zeiger))
      { start.unten *= 10;
        i *= 10;
        start.oben *= 10;
        start.unten += *zeiger++ - '0';
        i += start.unten >> 16;
        start.oben += i >> 16;
        start.unten &= 0xFFFFul;
        i &= 0xFFFFul;
      } /* ULTRALANG Eingabe Start */
      start.unten |= i << 16;
    case 3:
      zeiger = argv[--argc];
      anzahl = 0;
      while (isdigit(*zeiger))
      { anzahl *= 10;
        anzahl += *zeiger++ - '0';
      } /* Eingabe Anzahl */
    case 2:
      primfile = argv[1];
      break;
    default:
      puts("Syntax: primlist <Datei> [[Start] Anzahl]");
      exit(-1);
  } /* Kommandozeile */
  if ((datei = fopen(primfile, "rb")) == NULL)
  { printf("Fehler: kann Datei %s nicht ffnen.\n", primfile);
    exit(-3);
  }
  /* erst den unsigned 64 Bit Offset holen */
  offset.unten = (unsigned int) getw(datei);
  offset.unten |= (unsigned long int) (unsigned int) getw(datei) << 16;
  offset.oben = (unsigned int) getw(datei);
  offset.oben |= (unsigned long int) (unsigned int) getw(datei) << 16;
  if ((i = uldiff(start, offset)) == ~0ul ||
      fseek(datei, (i >>= 5) << 1, SEEK_CUR))
  { puts("Die Datei endet bereits vor oder "
         "beginnt erst hinter dem gewnschten Start.");
    printf("Gewnschter Start:");
    gibaus(start);
    printf("\nDatei beginnt bei:");
    gibaus(offset);
    putchar('\n');
    exit(-6);
  } /* Lesezeiger positioniert und i angepat */
  while (anzahl)
  { wort = getw(datei);
    if (feof(datei))
      break;
    if (ferror(datei))
    { fclose(datei);
      printf("\nFehler beim Lesen aus Datei %s.\n", primfile);
      exit(-4);
    } /* ferror */
    for (j = 0; j < 16; j++)
    { if (wort & 1u << j)
      { gibaus(uladd(offset, (((i<<4) | j) << 1) | 1));
        if (!--anzahl)
          break;
      } /* Primzahl */
    } /* Wort */
    i++;
  } /* Datei oder Anzahl */
  if (anzahl)
    puts("\nDateiende erreicht.");
  fclose(datei);
  putchar('\n');
  return 0;
} /* main */

unsigned long int uldiff(ULTRALANG ul1, ULTRALANG ul2)
{ ul2.unten = ul1.unten - ul2.unten;
  if (ul1.unten < ul2.unten) /* Unterlauf hat stattgefunden */
    ul1.oben--;
  if (ul1.oben == ul2.oben)
    return ul2.unten;
  else
    return ~0ul;
} /* uldiff */

ULTRALANG uladd(ULTRALANG ulzahl, unsigned long int zahl)
{ ulzahl.unten += zahl;
  if (ulzahl.unten < zahl) /* berlauf hat stattgefunden */
    ulzahl.oben++;
  return ulzahl;
} /* uladd */

void gibaus(ULTRALANG ulzahl)
{ static unsigned int s=0;  /* aktuelle Spalte */
  int i, rest[3];
  ldiv_t erg;

  if (ulzahl.oben)
  { i = 3;
    do
    { erg = ldiv(ulzahl.oben, 10000l);
      erg.rem = erg.rem << 16 | ulzahl.unten >> 16;
      ulzahl.oben = erg.quot;
      erg = ldiv(erg.rem, 10000l);
      erg.rem = erg.rem << 16 | ulzahl.unten & 0xFFFFul;
      ulzahl.unten = erg.quot << 16;
      erg = ldiv(erg.rem, 10000l);
      rest[--i] = (int) erg.rem;
      ulzahl.unten |= erg.quot;
    } while (i);
    if (ulzahl.unten)
    { if ((s += 19) > ZEILE)
      { putchar('\n');
        s = 19;
      } /* neue Zeile */
      printf("%7lu", ulzahl.unten);
      while (i < 3)
        printf("%04d", rest[i++]);
    } /* Rest und Restefeld */
    else
    { if ((s += 13) > ZEILE)
      { putchar('\n');
        s = 13;
      } /* neue Zeile */
      printf("%5d", rest[i++]);
      while (i < 3)
        printf("%04d", rest[i++]);
    } /* nur Restefeld */
  } /* ultralang */
  else
  { if ((s += 11) > ZEILE)
    { putchar('\n');
      s = 11;
    } /* neue Zeile */
    printf("%11lu", ulzahl.unten);
  } /* nur unteres long */
} /* gibaus */
