/* PRIMPROJ.C:
Anlegen einer Primflag-Datei <projfile> mit Offset <beginn> fr 
<bereich> unter Verwendung einer bestehenden Primflag-Datei <bisher> 
mit Offset 0. Dabei mu die bisherige Datei bis Wurzel <beginn> +
<bereich> gehen. Falls sie zuvor zuende ist, wird das Programm
abgebrochen.

Aufruf: primproj <bisher> <beginn> [<bereich> [<projfile>]] */ 


#include <stdio.h>
#include <ctype.h>
#include <malloc.h>

#define STDGR 100000ul

typedef struct
  { unsigned long int oben;
    unsigned long int unten;
  } ULTRALANG;

int befehl(int, char *[]);
int sieve(void);
ULTRALANG uladd(ULTRALANG, unsigned long int);
ULTRALANG ulmul(unsigned long int, unsigned long int);
unsigned long int uldiv(ULTRALANG, unsigned long int);
unsigned long int ulmod(ULTRALANG, unsigned long int);
int speichern(void);

unsigned long int anzahl = (STDGR - 1) / 32 + 1;
ULTRALANG beginn = {0L, 0L};
unsigned int _huge *primflag = NULL;
char *bisher, *projfile = "PRIMPROJ.BIT";
FILE *datei = NULL;


main(int argc, char *argv[])
{ int i;

  puts("PRIMPROJ 1.0 - Erzeugt abgeleitete (nicht bei 0 beginnende)");
  puts("               Primzahlendatei     (c) 1995 Heinz Repp");

  if (i = (befehl(argc, argv)))
    goto abbruch;

  if ((datei = fopen(bisher, "rb")) == NULL)
  { printf("Fehler: kann Datei %s nicht ffnen.\n", bisher);
    i = -3;
    goto abbruch;
  } /* Quelldatei geffnet */

  for (i = 4; i; i--)
  { if (getw(datei))
    { puts("Lesefehler oder Dateioffset nicht 0.");
      i = -6;
      goto abbruch;
    } /* mssen 0 sein: */
  } /* die ersten 4 Worte */

  if ((primflag = halloc(anzahl, sizeof(int))) == NULL)
  { printf("Ungengend Speicher fr %lu ", anzahl * 32);
    puts(" - Programm beendet.");
    i = -2;
    goto abbruch;
  } /* dynamisches Huge-Feld zugeordnet */

  if (i = sieve())
    goto abbruch;

  fclose(datei);

  if ((datei = fopen(projfile, "wb")) == NULL)
  { printf("\nFehler: kann Datei %s nicht ffnen - Abbruch.\n", projfile);
    i = -3;
    goto abbruch;
  } /* Datei zum Schreiben geffnet */

  i= speichern();

abbruch:
  if (primflag)
    hfree(primflag);
  if (datei)
    fclose(datei);
  exit(i);
} /* main */


int befehl(int argc, char *argv[])
{ unsigned long int zwischen;
  char *ziffer;

  switch (argc)
  { case 5:
      projfile = argv[4];
    case 4:
      sscanf(argv[3], "%lu", &anzahl);
      anzahl = (anzahl - 1) / 32 + 1;
    case 3:
      ziffer = argv[2];
      zwischen = 0l; /* Zwischenspeicher fr Bit 16-31 */
      while (isdigit(*ziffer))
      { beginn.unten *= 10;
        zwischen *= 10;
        beginn.oben *= 10;
        beginn.unten += *ziffer++ - '0';
        zwischen += beginn.unten >> 16;
        beginn.oben += zwischen >> 16;
        beginn.unten &= 0xFFFFul;
        zwischen &= 0xFFFFul;
      } /* ULTRALANG Eingabe Beginn */
      beginn.unten |= zwischen << 16;
      beginn.unten &= ~31ul; /* mu durch 32 teilbar sein */

      bisher = argv[1];
      break;
    default:
      puts("Syntax: primproj <Basisdatei> <Beginn> "
           "[<Bereich> [<Ausgabedatei>]]");
      return -1;
  } /* Weichensteller */
  return 0;
} /* befehl */


int sieve(void)
{ unsigned long int i, prim, zahl, ende;
  unsigned int j, count, flagwort;
  ULTRALANG enorm;

  printf("rechne ");
  /* Flagfeld mit Einsen initialisieren */
  for (i = 0; i < anzahl; primflag[i++] = ~0u)
    ;

  zahl = 0x80000000ul;
  enorm.oben = beginn.oben >> 5;
  enorm.unten = beginn.unten >> 5 | beginn.oben << 27;
  enorm = uladd(enorm, anzahl);
  enorm.unten = enorm.unten >> 5 | enorm.oben << 27;
  enorm.oben = enorm.oben >> 5;
  do
  { ende = zahl;
    zahl = (uldiv(enorm, ende) + ende) / 2;
  } while (ende > zahl);


  enorm.oben = beginn.oben >> 1;
  enorm.unten = beginn.unten >> 1 | beginn.oben << 31;
  count = 1;
  for (i = 0; i <= ende; i++)
  { flagwort = getw(datei);

    if (feof(datei))
    { puts("\nunerwartetes Dateiende - Programm abgebrochen.");
      return -7;
    } /* EOF */

    if (ferror(datei))
    { puts("\nLesefehler - Programm abgebrochen.");
      return -4;
    } /* ferror */

    if (--count == 0)
    { putchar('.');
      count = uldiv(ulmul(i, ende), (40 * i + ende + 1)) + 1;
    } /* einen Punkt ausgeben */

    for (j = 0; j < 16; j++)
    { if (flagwort & (1u << j))
      { zahl = (i << 4) + j;
        prim = (zahl << 1) + 1;  /* jede 2. berspringen */
        zahl = prim - 1 - ulmod(uladd(enorm, zahl), prim);
        while (zahl < (anzahl << 4))
        { primflag[zahl >> 4] &= ~(1 << (zahl&15));
          zahl += prim;
        } /* Lschschleife */
      } /* endif Bit gesetzt */
    } /* Bitschleife */

  } /* Wortschleife */
  puts(" fertig.");
  return 0;
} /* sieve */


ULTRALANG uladd(ULTRALANG ulzahl, unsigned long int zahl)
{ ulzahl.unten += zahl;
  if (ulzahl.unten < zahl) /* berlauf hat stattgefunden */
    ulzahl.oben++;
  return ulzahl;
} /* uladd */


ULTRALANG ulmul(unsigned long int zahl1, unsigned long int zahl2)
{ ULTRALANG erg;
  unsigned long int zwischen;
  erg.unten = (zahl1 & 0xFFFFul) * (zahl2 & 0xFFFFul);
  erg.oben = (zahl1 & 0xFFFFul) * (zahl2 >> 16) + (erg.unten >> 16);
  erg.unten = (erg.unten & 0xFFFFul) | (erg.oben << 16);
  zwischen = erg.oben >> 16;
  erg.oben = (zahl1 >> 16) * (zahl2 & 0xFFFFul) + (erg.unten >> 16);
  erg.unten = (erg.unten & 0xFFFFul) | (erg.oben << 16);
  zwischen += erg.oben >> 16;
  erg.oben = (zahl1 >> 16) * (zahl2 >> 16) + zwischen;
  return erg;
} /* ulmul */


/* Divisions- und Modulofunktion setzen voraus, da das hchstwertige Bit
der ulzahl gelscht und die zahl nicht grer als 0x80000000ul ist. */

unsigned long int uldiv(ULTRALANG ulzahl, unsigned long int zahl)
{ unsigned long int ergebnis;
  int i;
  ergebnis = 0l;
  for (i = 31; i>=0; i--)
  { ulzahl.oben = ulzahl.oben << 1 | ulzahl.unten >> 31;
    ulzahl.unten <<= 1;
    if (ulzahl.oben >= zahl)
    { ulzahl.oben -= zahl;
      ergebnis |= 1ul << i;
    } /* endif */
  } /* endfor */
  return ergebnis;
} /* uldiv */


unsigned long int ulmod(ULTRALANG ulzahl, unsigned long int zahl)
{ int i;
  ulzahl.oben %= zahl;
  for (i = 32; i; i--)
  { ulzahl.oben = ulzahl.oben << 1 | ulzahl.unten >> 31;
    ulzahl.unten <<= 1;
    if (ulzahl.oben >= zahl)
    { ulzahl.oben -= zahl;
    } /* endif */
  } /* endfor */
  return ulzahl.oben;
} /* ulmod */


int speichern(void)
{ unsigned long int i;

  putw((unsigned int) (beginn.unten & 0xFFFFl), datei);
  putw((unsigned int) (beginn.unten >> 16), datei);
  putw((unsigned int) (beginn.oben & 0xFFFFl), datei);
  putw((unsigned int) (beginn.oben  >> 16), datei);

  for (i = 0; i < anzahl; )
  { putw(primflag[i++], datei);
    if (ferror(datei))
    { printf("\nFehler beim Schreiben in %s:\n", projfile);
      puts("z.B. Datentrger voll oder Datei schreibgeschtzt?");
      return -5;
    } /* Schreibfehler */
  } /* Feld gespeichert */
  return 0;
} /* speichern */
