/* PRIMTEST.C - Test, ob Zahl prim ist, wenn nein Ausgabe der Faktoren
Aufruf: primtest <datei> <zahl>, beide Paramenter sind obligat */


#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>

typedef struct
  { unsigned long int oben;
    unsigned long int unten;
  } ULTRALANG;

void gibaus(ULTRALANG, int);
long int ulsqrt(ULTRALANG);
ULTRALANG uldiv(ULTRALANG, long int);
long int ulmod(ULTRALANG, long int);


int main(int argc, char *argv[])
{ int prim;
  unsigned int flags, maske;
  long int inter, ende;
  char *ziffer;
  FILE *datei;
  ULTRALANG zahl = {0L, 0L};

  puts("PRIMTEST 1.0 - Primfaktorenzerlegung (c) 1995 Heinz Repp");
  if (argc != 3)
  { puts("Syntax: primtest <Basisdatei> <Zahl>");
    exit(-1);
  } /* Syntaxfehler */
  else
  { ziffer = argv[2];
    inter = 0L; /* Zwischenspeicher fr Bit 16-31 */
    while (isdigit(*ziffer))
    { zahl.unten *= 10;
      inter *= 10;
      zahl.oben *= 10;
      zahl.unten += *ziffer++ - '0';
      inter += zahl.unten >> 16;
      zahl.oben += inter >> 16;
      zahl.unten &= 0xFFFFul;
      inter &= 0xFFFFul;
    } /* ULTRALANG Zahl */
    zahl.unten |= inter << 16;
  } /* endif */

  if ((datei = fopen(argv[1], "rb")) == NULL)
  { printf("Fehler: kann Datei %s nicht ffnen.\n", argv[1]);
    exit(-3);
  } /* Quelldatei geffnet */

  for (flags = 4; flags; flags--)
  { if (getw(datei))
    { puts("Lesefehler oder Dateioffset nicht 0.");
      fclose(datei);
      exit(-6);
    } /* mssen 0 sein: */
  } /* die ersten 4 Worte */

  prim = -1;
  flags = ~0;
  maske = 1 << 15;
  ende = ulsqrt(zahl);
  inter = 2L;
  while (inter <= ende)
  { if (flags & maske)
    { while (ulmod(zahl, inter) == 0L)
      { if (prim)
        { printf("Die Zahl ");
          gibaus(zahl, 0);
          puts(" ist keine Primzahl.");
          puts("Sie hat folgende Primfaktoren:");
          prim = 0;
        } /* endif */
        printf("%19ld\n", inter);
        zahl = uldiv(zahl, inter);
        ende = ulsqrt(zahl);
      } /* endwhile */
      if (inter == 2L)
        inter = -1L;
    } /* endif */


    if (!(maske <<= 1))
    { flags = getw(datei);
      if (feof(datei))
      { puts("unerwartetes Dateiende - Programm abgebrochen.");
        fclose(datei);
        exit(-7);
      } /* EOF */
      if (ferror(datei))
      { puts("Lesefehler - Programm abgebrochen.");
        fclose(datei);
        exit(-4);
      } /* ferror */
      maske = 1;
    } /* ggf. nchstes Wort gelesen */

    inter += 2;
  } /* Bitschleife */

  fclose(datei);
  if (prim)
  { printf("Die Zahl ");
    gibaus(zahl, 0);
    puts(" ist eine Primzahl.");
    exit(1);
  }
  else
  { gibaus(zahl, -1);
    printf("\n");
    exit(0);
  }
} /* main */


void gibaus(ULTRALANG ulzahl, int flag)
{ int i, rest[3];
  ldiv_t erg;

  if (ulzahl.oben)
  { i = 3;
    do
    { erg = ldiv(ulzahl.oben, 10000l);
      erg.rem = erg.rem << 16 | ulzahl.unten >> 16;
      ulzahl.oben = erg.quot;
      erg = ldiv(erg.rem, 10000l);
      erg.rem = erg.rem << 16 | ulzahl.unten & 0xFFFFul;
      ulzahl.unten = erg.quot << 16;
      erg = ldiv(erg.rem, 10000l);
      rest[--i] = (int) erg.rem;
      ulzahl.unten |= erg.quot;
    } while (i);
    if (ulzahl.unten)
    { printf("%*lu", flag ? 7 : 0, ulzahl.unten);
      while (i < 3)
        printf("%04d", rest[i++]);
    } /* Rest und Restefeld */
    else
    { printf("%*d", flag ? 11 : 0, rest[i++]);
      while (i < 3)
        printf("%04d", rest[i++]);
    } /* nur Restefeld */
  } /* ultralang */
  else
  { printf("%*lu", flag ? 19 : 0, ulzahl.unten);
  } /* nur unteres long */
} /* gibaus */


long int ulsqrt(ULTRALANG ulzahl)
{ long int wurzel;
  ULTRALANG merke = {0L, 0x7FFFFFFFul};
  do 
  { wurzel = merke.unten;
    merke = uldiv(ulzahl, wurzel);
    merke.unten = (merke.unten + (unsigned long int) wurzel) / 2;
  } while (wurzel > merke.unten);
  return wurzel;
} /* ulsqrt */


ULTRALANG uldiv(ULTRALANG ulzahl, long int zahl)
{ ULTRALANG ergebnis = {0L, 0L};
  unsigned long int akku = 0L;
  int i;
  for (i = 31; i>=0; i--)
  { akku = akku << 1 | ulzahl.oben >> 31;
    ulzahl.oben <<= 1;
    if (akku >= zahl)
    { akku -= zahl;
      ergebnis.oben |= 1ul << i;
    } /* endif */
  } /* endfor */
  for (i = 31; i>=0; i--)
  { akku = akku << 1 | ulzahl.unten >> 31;
    ulzahl.unten <<= 1;
    if (akku >= zahl)
    { akku -= zahl;
      ergebnis.unten |= 1ul << i;
    } /* endif */
  } /* endfor */
  return ergebnis;
} /* uldiv */


long int ulmod(ULTRALANG ulzahl, long int zahl)
{ register unsigned long int akku, gleit;
  register int i;
  akku = ulzahl.oben % zahl;
  gleit = ulzahl.unten;
  for (i = 32; i; i--)
  { akku = akku << 1 | gleit >> 31;
    gleit <<= 1;
    if (akku >= zahl)
    { akku -= zahl;
    } /* endif */
  } /* endfor */
  return akku;
} /* ulmod */
