/* Primzahlensuche mit dem Sieb des Erastothenes:
Die Menge der Primzahlen besteht anfangs aus allen natrlichen Zahlen 
auer 1. Beginnend bei 2 werden nun fr jedes noch enthaltene Element 
alle Vielfachen davon aus der Menge entfernt. Dadurch sind jeweils 
ausschlielich bis zum Quadrat des nchsten Elements nur mehr Zahlen 
enthalten, die nicht durch eine kleinere Zahl teilbar und daher prim 
sind. Da beim ersten Schritt bereits alle geraden Zahlen fortfallen, 
brauchen diese im weiteren nicht mehr bercksichtige werden. Fr die 
ungeraden Zahlen wird ein Bitfeld angelegt, wobei ein gesetztes Bit 
bedeutet, da die Zahl noch enthalten ist. Die Adresse des 
entsprechenden Bits ist (Zahl - 1) / 2, binr eine Rechtsverschiebung 
um ein Bit. Ein Byte steht fr 8 (ungerade von insgesamt 16) Zahlen 
(1,3,...,13,15). Die Zahl bestimmt daher wie folgt die Adresse des 
Bytes und des entsprechenden Bits:  

Byteadresse = Zahl>>4;              (Bits 4ff)
Position im Byte = (Zahl&15)>>1;    (Bits 1-3)

Praktischerweise sollte immer ein ganzes Wort gelesen und darin die
einzelnen Bits getestet werden:

Wortadresse = Zahl>>5;
Position im Wort = (Zahl&31)>>1;  */


/* Kommandozeilenparameter:
Zahl: Primzahlsuche bis 1000*Zahl, Standard STDGR
Filename: Datei, unter der das Bitfeld abgespeichert werden soll,
Standard primxxxx.bit, wobei xxxx=Bereich/2000 */


/* Fehlercodes (gilt fr alle Programme:
   1 = Zahl ist Primzahl (nur PRIMTEST)
   0 = kein Fehler
  -1 = Syntaxfehler
  -2 = nicht gengend Speicher fr Bitfeld
  -3 = Fehler bei Dateiffnen
  -4 = Lesefehler
  -5 = Schreibfehler
  -6 = fehlerhafter Dateioffset
  -7 = vorzeitiges Dateiende */


#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <malloc.h>

#define STDGR 100000ul

/* Bitfeld und Gre sind globale Gren */
unsigned long int anzahl = (STDGR - 1) / 32 + 1;
unsigned int _huge *primflag;

void sieve(void);
int speichern(char *);


main(int argc, char *argv[])
{ char primfile[128]="";
  int i;

  puts("PRIMZAHL 1.0 - Erzeugt primre Primzahlendatei (c) 1995 Heinz Repp");
  for (i=1; i<argc; i++)
  { if (isdigit(*argv[i]))
    { sscanf(argv[i], "%lu", &anzahl);
      anzahl = (anzahl - 1) / 32 + 1;
    } /* Gre des Feldes angepat */
    else
      strncpy(primfile, argv[i], 127);
  } /* Kommandozeile ausgewertet */
  if (*primfile == '\0')
    sprintf(primfile, "PRIM%04lu.BIT", anzahl * 2 / 125);
  if ((primflag = halloc(anzahl, sizeof(int))) == NULL)
  { printf("Ungengend Speicher fr Bereich bis %lu", anzahl * 32);
    puts(" - Programm beendet.");
    exit(-2);
  } /* dynamisches Huge-Feld zugeordnet */
  sieve();
  if (i = speichern(primfile))
  { printf("Fehler beim Speichern von %s:\n", primfile);
    puts("z.B. Datentrger voll oder Datei schreibgeschtzt?");
    exit(i);
  } /* Feld gespeichert */
  hfree(primflag);
  exit(0);
} /* main */


void sieve(void)
{ unsigned long int i, prim, zahl, ende;
  unsigned int j, count;
  printf("rechne ");
  /* Flagfeld mit Einsen initialisieren bis auf 1 */
  primflag[0] = ~1u;  /* damit 1 nicht bercksichtigt wird */
  for (i = 0; i < anzahl; primflag[++i] = ~0u)
    ;
  count = 1;        /* so gibts gleich einen Punkt */
  zahl = (anzahl >> 6);
  do
  { ende = zahl;
    zahl = ((anzahl >> 5) / ende + ende) / 2;
  } while (ende > zahl);

  for (i = 0; i <= ende; i++)
  { if (--count == 0)
    { putchar('.');
      count = i * ende / (40 * i + ende + 1) + 1;
    } /* einen Punkt ausgeben */
    for (j = 0; j < 16; j++)
    { if (primflag[i] & 1u << j)
      { zahl = (i << 4) + j;
        prim = (zahl << 1) + 1;  /* jede 2. berspringen */
        while ((zahl += prim) < anzahl << 4)
          primflag[zahl >> 4] &= ~(1 << (zahl&15));
      } /* endif Bit gesetzt */
    } /* Bitschleife */
  } /* Wortschleife */
  puts(" fertig.");
} /* sieve */


int speichern(char *dateiname)
{ FILE *datei;
  unsigned long int i;
  if ((datei = fopen(dateiname, "wb")) == NULL)
    return -3;
  /* am Anfang ein unsigned 64 Bit Offset, hier 0 */
  for (i = 4; i; i--)
    putw(0, datei);
  /* dann das Bitfeld */
  while(i < anzahl)
  { putw(primflag[i++], datei);
    if (ferror(datei))
    { fclose(datei);
      return -5;
    } /* ferror */
  } /* Feld gespeichert */
  if (fclose(datei))
    return -1;
  return 0;
} /* speichern */
